﻿using System;
using System.Threading.Tasks;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Linq;

namespace metastrings
{
    public class CsvParser : IDisposable
    {
        public CsvParser(Stream stream, bool keepOpen = false)
            : this(new StreamReader(stream, Encoding.UTF8), keepOpen)
        {
        }

        public CsvParser(TextReader reader, bool keepOpen = false)
        {
            m_tr = reader;
            m_keepOpen = keepOpen;
        }

        public void Dispose()
        {
            if (m_tr != null && !m_keepOpen)
            {
                m_tr.Dispose();
                m_tr = null;
            }
        }

        public bool Read()
        {
            string line;
            while (true)
            {
                line = m_tr.ReadLine();
                if (line == null)
                    return false;

                if (string.IsNullOrWhiteSpace(line))
                    continue;

                break;
            }

            ParseCsvLine(line, m_sb, m_lineBuffer);
            return true;
        }

        public async Task<bool> ReadAsync()
        {
            string line;
            while (true)
            {
                line = await m_tr.ReadLineAsync().ConfigureAwait(false);
                if (line == null)
                    return false;

                if (string.IsNullOrWhiteSpace(line))
                    continue;

                break;
            }

            ParseCsvLine(line, m_sb, m_lineBuffer);
            return true;
        }

        public int Count => m_lineBuffer.Count;
        public string this[int index] => m_lineBuffer[index];

        public static void ParseCsvLine(string line, StringBuilder sb, List<string> lineBuffer)
        {
            sb.Clear();
            lineBuffer.Clear();

            if (line.Length == 0)
                return;

            bool hasQuotes = line.IndexOf('\"') >= 0;
            if (!hasQuotes)
            {
                lineBuffer.AddRange(line.Split(',').Select(s => CsvPrep.UnNullCsv(s)));
                return;
            }

            bool inQuote = false;
            for (int c = 0; c < line.Length; ++c)
            {
                char cur = line[c];

                if (cur == '\"')
                {
                    if (sb.Length == 0) // start quote
                    {
                        inQuote = true;
                        continue;
                    }

                    // quote-before double
                    if (c + 1 < line.Length)
                    {
                        char next = line[c + 1];
                        if (next == '\"')
                        {
                            sb.Append('\"');
                            ++c;
                            continue;
                        }
                    }

                    // end quote
                    string cell = sb.ToString();
                    lineBuffer.Add(CsvPrep.UnNullCsv(sb.ToString()));
                    sb.Clear();
                    ++c; // skip comma
                    inQuote = false;
                    continue;
                }

                if (cur == ',' && !inQuote)
                {
                    lineBuffer.Add(CsvPrep.UnNullCsv(sb.ToString()));
                    sb.Clear();
                    continue;
                }

                sb.Append(cur);
            }

            if (sb.Length > 0 || line[line.Length - 1] == ',')
                lineBuffer.Add(CsvPrep.UnNullCsv(sb.ToString()));
        }

        private TextReader m_tr;
        private bool m_keepOpen;
        private StringBuilder m_sb = new StringBuilder();
        private List<string> m_lineBuffer = new List<string>();
    }
}
