﻿using System;
using System.Collections.Generic;
using System.Threading.Tasks;
using System.Text;
using System.IO;

using MySql.Data.MySqlClient;

namespace metastrings
{
    public class Command : IDisposable
    {
        public Command(Context ctxt)
        {
            Ctxt = ctxt;
        }

        public void Dispose()
        {
            if (Cmd != null)
            {
                Cmd.Dispose();
                Cmd = null;
            }
        }

        public async Task DefineAsync(Define define)
        {
            var totalTimer = ScopeTiming.StartTiming();
            try
            {
                var localTimer = ScopeTiming.StartTiming();

                if (define.metadata == null || define.metadata.Count == 0)
                    return;

                var values = new List<object>(define.metadata.Count);
                values.AddRange(define.metadata.Keys); // keys are the values of items with metadata

                bool firstValueIsNumeric = !(values[0] is string);
                short typeId = await Types.GetIdAsync(Ctxt, define.type, firstValueIsNumeric).ConfigureAwait(false);
                TypeObj type = await Types.GetTypeAsync(Ctxt, typeId).ConfigureAwait(false);

                foreach (object value in values)
                {
                    bool valueIsNumeric = !(value is string);
                    if (type.isNumeric != valueIsNumeric)
                        throw new MetaStringsException($"Type value numeric does not match type numeric: {define.type}");
                }
                ScopeTiming.RecordScope("Define.Setup", localTimer);

                Ctxt.DbConn.BeginTrans();

                var valueIdCache = new Dictionary<object, long>(values.Count); // value => valueid
                foreach (object value in values)
                    valueIdCache[value] = await Values.GetIdAsync(Ctxt, value).ConfigureAwait(false);
                ScopeTiming.RecordScope("Define.Values", localTimer);

                var valueIdsItemIdCache = new Dictionary<object, long>(values.Count); // value => itemid
                foreach (object value in values)
                {
                    long itemId = await Items.GetIdAsync(Ctxt, typeId, valueIdCache[value]).ConfigureAwait(false);
                    valueIdsItemIdCache[value] = itemId;
                }
                ScopeTiming.RecordScope("Define.Items", localTimer);

                var nameIds = new Dictionary<string, short>();
                foreach (object value in values)
                {
                    var metadata = define.metadata[value];
                    if (metadata == null || metadata.Count == 0)
                        continue;

                    foreach (var kvp in metadata)
                    {
                        short nameId;
                        if (!nameIds.TryGetValue(kvp.Key, out nameId))
                        {
                            bool isValueNumeric = !(kvp.Value is string);
                            nameId = await Names.GetIdAsync(Ctxt, typeId, kvp.Key, isValueNumeric).ConfigureAwait(false);
                            nameIds.Add(kvp.Key, nameId);
                        }
                    }
                }
                ScopeTiming.RecordScope("Define.NameIds", localTimer);

                var nameObjs = new Dictionary<short, NameObj>();
                foreach (short nameId in nameIds.Values)
                {
                    NameObj nameObj;
                    if (!nameObjs.TryGetValue(nameId, out nameObj))
                    {
                        nameObj = await Names.GetNameAsync(Ctxt, nameId).ConfigureAwait(false);
                        nameObjs.Add(nameId, nameObj);
                    }
                }
                ScopeTiming.RecordScope("Define.NameObjs", localTimer);

                foreach (object value in values)
                {
                    var metadata = define.metadata[value];
                    if (metadata == null || metadata.Count == 0)
                        continue;

                    var nameValueIds = new Dictionary<short, long>();
                    foreach (var kvp in metadata)
                    {
                        bool isValueNumeric = !(kvp.Value is string);

                        short nameId = nameIds[kvp.Key];
                        if (kvp.Value == null)
                        {
                            nameValueIds[nameId] = -1;
                            continue;
                        }

                        // Type safety
                        var nameObj = nameObjs[nameId];
                        if (isValueNumeric != nameObj.isNumeric)
                            throw new MetaStringsException($"Name value numeric does not match name numeric: {nameObj.name}");

                        if (kvp.Value == null)
                        {
                            nameValueIds[nameId] = -1;
                        }
                        else if (!string.IsNullOrEmpty(kvp.Value.ToString()))
                        {
                            long valueId;
                            if (!valueIdCache.TryGetValue(kvp.Value, out valueId))
                            {
                                valueId = await Values.GetIdAsync(Ctxt, kvp.Value).ConfigureAwait(false);
                                valueIdCache[kvp.Value] = valueId;
                            }

                            nameValueIds[nameId] = valueId;
                        }
                    }

                    Items.SetItemData(Ctxt, valueIdsItemIdCache[value], nameValueIds);
                }
                Ctxt.DbConn.Commit();
                ScopeTiming.RecordScope("Define.Metadata", localTimer);

                await Ctxt.ProcessPostOpsAsync().ConfigureAwait(false);
                ScopeTiming.RecordScope("Define.Complete", localTimer);
            }
#if !DEBUG
            catch
            {
                Ctxt.ClearPostItemOps();
                NameValues.ClearCaches();
                throw;
            }
#endif
            finally
            {
                ScopeTiming.RecordScope("Define", totalTimer);
            }
        }

        public async Task DefineCsvAsync(string csv)
        {
            var totalTimer = ScopeTiming.StartTiming();
            try
            {
                var localTimer = ScopeTiming.StartTiming();

                Ctxt.DbConn.BeginTrans();

                using (CsvParser parser = new CsvParser(new StringReader(csv)))
                {
                    short typeId;
                    {
                        if (!await parser.ReadAsync().ConfigureAwait(false) || parser.Count < 1 || parser[0] != "type")
                            throw new MetaStringsException($"The first line is for header names");
                        if (!await parser.ReadAsync().ConfigureAwait(false) || parser.Count < 1)
                            throw new MetaStringsException($"The second line is for the header values, like the table name");
                        string typeName = parser[0];
                        typeId = await Types.GetIdAsync(Ctxt, typeName).ConfigureAwait(false);
                    }

                    if (!await parser.ReadAsync().ConfigureAwait(false))
                        throw new MetaStringsException($"The third line is for data columns, with the first column reserved for the item value, aka primary key");

                    List<short> nameIds = new List<short>(parser.Count);
                    nameIds.Add(-1); // start off with the item value column, un-named

                    bool[] isNumericColumn = new bool[parser.Count];
                    isNumericColumn[0] = false;
                    
                    for (int s = 1; s < parser.Count; ++s)
                    {
                        string columnName = parser[s];
                        
                        bool isNumeric = columnName.EndsWith("#", StringComparison.Ordinal);
                        if (isNumeric)
                            columnName = columnName.Substring(0, columnName.Length - 1);
                        isNumericColumn[s] = isNumeric;

                        short nameId = await Names.GetIdAsync(Ctxt, typeId, columnName, isNumeric).ConfigureAwait(false);
                        nameIds.Add(nameId);
                    }
                    ScopeTiming.RecordScope("DefineCsv.Setup", localTimer);

                    Dictionary<short, long> itemMetadata = new Dictionary<short, long>();

                    Dictionary<string, long> valueIdCache = new Dictionary<string, long>();
                    int rowNumber = 1;
                    while (await parser.ReadAsync().ConfigureAwait(false))
                    {
                        itemMetadata.Clear();

                        if (parser.Count != nameIds.Count)
                            throw new MetaStringsException($"Row #{rowNumber} has {parser.Count} columns, but there were {nameIds.Count} column headers");

                        long itemId;
                        {
                            long itemValueId = await Values.GetIdAsync(Ctxt, parser[0]).ConfigureAwait(false);
                            itemId = await Items.GetIdAsync(Ctxt, typeId, itemValueId).ConfigureAwait(false);
                        }

                        for (int v = 1; v < parser.Count; ++v)
                        {
                            long valueId = -1;
                            {
                                string curStringValue = parser[v];
                                if (curStringValue.Length > 0)
                                {
                                    if (!valueIdCache.TryGetValue(curStringValue, out valueId))
                                    {
                                        if (isNumericColumn[v])
                                        {
                                            double numValue;
                                            if (!double.TryParse(curStringValue, out numValue))
                                                throw new MetaStringsException($"Row #{rowNumber} column #{v + 1} does not parse as number");
                                            valueId = await Values.GetIdAsync(Ctxt, numValue);
                                        }
                                        else
                                            valueId = await Values.GetIdAsync(Ctxt, curStringValue);

                                        valueIdCache[curStringValue] = valueId;
                                    }
                                }
                            }

                            if (valueId >= 0)
                            {
                                short nameId = nameIds[v];
                                itemMetadata[nameId] = valueId;
                            }
                        }

                        Items.SetItemData(Ctxt, itemId, itemMetadata);

                        ++rowNumber;
                    }
                    Ctxt.DbConn.Commit();
                    ScopeTiming.RecordScope("DefineCsv.Metadata", localTimer);

                    await Ctxt.ProcessPostOpsAsync().ConfigureAwait(false);
                    ScopeTiming.RecordScope("DefineCsv.Complete", localTimer);
                }
            }
#if !DEBUG
            catch
            {
                Ctxt.ClearPostItemOps();
                NameValues.ClearCaches();
                throw;
            }
#endif
            finally
            {
                ScopeTiming.RecordScope("DefineCsv", totalTimer);
            }
        }

        public async Task<string> GenerateSqlAsync(Select query)
        {
            var totalTimer = ScopeTiming.StartTiming();
            try
            {
                string sql = await Sql.GenerateSqlAsync(Ctxt, query).ConfigureAwait(false);
                return sql;
            }
            finally
            {
                ScopeTiming.RecordScope("Cmd.GenerateSql", totalTimer);
            }
        }

        public async Task<Reader> GetReaderAsync(Select query)
        {
            var totalTimer = ScopeTiming.StartTiming();
            try
            {
                string sql = await GenerateSqlAsync(query).ConfigureAwait(false);
                return await GetReaderAsync(sql, query.cmdParams).ConfigureAwait(false);
            }
            finally
            {
                ScopeTiming.RecordScope("Cmd.GenerateSql(query)", totalTimer);
            }
        }

        public async Task<Reader> GetReaderAsync(string sql, Dictionary<string, object> cmdParams)
        {
            var totalTimer = ScopeTiming.StartTiming();
            try
            {
                Cmd = new MySqlCommand(sql, Ctxt.DbConn.MyDb, Ctxt.DbConn.MyTrans);

                if (cmdParams != null)
                {
                    foreach (var kvp in cmdParams)
                        Cmd.Parameters.AddWithValue(kvp.Key, kvp.Value);
                }

                var myReader = await Cmd.ExecuteReaderAsync().ConfigureAwait(false);
                Reader reader = new Reader(myReader);
                return reader;
            }
            finally
            {
                ScopeTiming.RecordScope("Cmd.GenerateSql(sql)", totalTimer);
            }
        }

        public async Task<GetResponse> GetAsync(GetRequest request)
        {
            var totalTimer = ScopeTiming.StartTiming();
            try
            {
                var responses = new List<Dictionary<string, object>>(request.values.Count);

                short typeId = await Types.GetIdAsync(Ctxt, request.type, noCreate: true).ConfigureAwait(false);
                foreach (var value in request.values)
                {
                    long valueId = await Values.GetIdAsync(Ctxt, value).ConfigureAwait(false);

                    long itemId = await Items.GetIdAsync(Ctxt, typeId, valueId, noCreate: true).ConfigureAwait(false);
                    if (itemId < 0)
                    {
                        responses.Add(null);
                        continue;
                    }

                    var metaIds = await Items.GetItemDataAsync(Ctxt, itemId).ConfigureAwait(false);
                    var metaStrings = await NameValues.GetMetadataValuesAsync(Ctxt, metaIds).ConfigureAwait(false);

                    responses.Add(metaStrings);
                }

                GetResponse response = new GetResponse() { metadata = responses };
                return response;
            }
            finally
            {
                ScopeTiming.RecordScope("Cmd.Get", totalTimer);
            }
        }

        public async Task<GetResponse> QueryGetAsync(QueryGetRequest request)
        {
            var totalTimer = ScopeTiming.StartTiming();
            try
            {
                Select select = new Select();
                select.select = new List<string> { "id", "value" };
                select.type = request.type;
                select.where = request.where;
                select.orderBy = request.orderBy;
                select.limit = request.limit;
                select.cmdParams = request.cmdParams;

                var items = new List<KeyValuePair<long, string>>();
                using (var reader = await GetReaderAsync(select).ConfigureAwait(false))
                {
                    while (await reader.ReadAsync().ConfigureAwait(false))
                        items.Add(new KeyValuePair<long, string>(reader.GetInt64(0), reader.GetString(1)));
                }

                var responses = new List<Dictionary<string, object>>(items.Count);

                foreach (var kvp in items)
                {
                    var metaIds = await Items.GetItemDataAsync(Ctxt, kvp.Key).ConfigureAwait(false);
                    var metaStrings = await NameValues.GetMetadataValuesAsync(Ctxt, metaIds).ConfigureAwait(false);

                    metaStrings["value"] = kvp.Value;
                    responses.Add(metaStrings);
                }

                GetResponse response = new GetResponse() { metadata = responses };
                return response;
            }
            finally
            {
                ScopeTiming.RecordScope("Cmd.QueryGet", totalTimer);
            }
        }

        public async Task DeleteAsync(Delete toDelete)
        {
            var totalTimer = ScopeTiming.StartTiming();
            try
            {
                int typeId = await Types.GetIdAsync(Ctxt, toDelete.type, noCreate: true).ConfigureAwait(false);
                foreach (var val in toDelete.values)
                {
                    long valueId = await Values.GetIdAsync(Ctxt, val).ConfigureAwait(false);
                    string sql = $"DELETE FROM items WHERE valueid={valueId} AND typeid={typeId}";
                    Ctxt.AddPostItemOp(sql);
                }
            }
            finally
            {
                ScopeTiming.RecordScope("Cmd.Delete", totalTimer);
            }
        }

        public async Task DropAsync(Drop drop)
        {
            var totalTimer = ScopeTiming.StartTiming();
            try
            {
                int typeId = await Types.GetIdAsync(Ctxt, drop.type, noCreate: true).ConfigureAwait(false);

                using (var cmd = new MySqlCommand($"DELETE FROM itemnamevalues WHERE nameid IN (SELECT id FROM names WHERE typeid = {typeId})", Ctxt.DbConn.MyDb, Ctxt.DbConn.MyTrans))
                    await cmd.ExecuteNonQueryAsync().ConfigureAwait(false);

                using (var cmd = new MySqlCommand($"DELETE FROM names WHERE typeid = {typeId}", Ctxt.DbConn.MyDb, Ctxt.DbConn.MyTrans))
                    await cmd.ExecuteNonQueryAsync().ConfigureAwait(false);

                using (var cmd = new MySqlCommand($"DELETE FROM items WHERE typeid = {typeId}", Ctxt.DbConn.MyDb, Ctxt.DbConn.MyTrans))
                    await cmd.ExecuteNonQueryAsync().ConfigureAwait(false);

                using (var cmd = new MySqlCommand($"DELETE FROM types WHERE id = {typeId}", Ctxt.DbConn.MyDb, Ctxt.DbConn.MyTrans))
                    await cmd.ExecuteNonQueryAsync().ConfigureAwait(false);
            }
            finally
            {
                ScopeTiming.RecordScope("Cmd.Drop", totalTimer);
            }
        }

        public void Reset(Reset reset)
        {
            if (reset.includeNameValues)
                NameValues.Reset(Ctxt);
            else
                Items.Reset(Ctxt);
        }

        private Context Ctxt;
        private MySqlCommand Cmd;
    }
}
