﻿using System;
using System.Collections.Generic;
using System.Collections.Concurrent;
using System.Threading.Tasks;
using System.Linq;
using System.Data;

using MySql.Data.MySqlClient;

namespace metastrings
{
    public class NameObj
    {
        public short id;
        public short typeId;
        public string name;
        public bool isNumeric;
    }

    public static class Names
    {
        public static void Reset(Context ctxt)
        {
            sm_cache.Clear();
            sm_cacheBack.Clear();

            using (var cmd = new MySqlCommand("DELETE FROM names", ctxt.DbConn.MyDb, ctxt.DbConn.MyTrans))
                cmd.ExecuteNonQuery();
        }

        public static async Task<short> GetIdAsync(Context ctxt, short typeId, string name, bool isNumeric = false, bool noCreate = false, bool noException = false)
        {
            var totalTimer = ScopeTiming.StartTiming();
            try
            {
                short id;
                string cacheKey = $"{typeId}_{name}";
                if (sm_cache.TryGetValue(cacheKey, out id))
                    return id;

                Exception lastExp = null;
                bool isExpFinal = false;
                for (int tryCount = 1; tryCount <= 4; ++tryCount)
                {
                    try
                    {
                        using (var cmd = new MySqlCommand("GetNameId", ctxt.DbConn.MyDb, ctxt.DbConn.MyTrans))
                        {
                            cmd.CommandType = CommandType.StoredProcedure;
                            cmd.Parameters.AddWithValue("typeIdParam", typeId);
                            cmd.Parameters.AddWithValue("nameParam", name);
                            cmd.Parameters.AddWithValue("isNumericParam", isNumeric);
                            cmd.Parameters.AddWithValue("noCreateParam", noCreate);
                            object idObj = await cmd.ExecuteScalarAsync().ConfigureAwait(false);
                            id = Convert.ToInt16(idObj);
                            if (id >= 0)
                            {
                                sm_cache[cacheKey] = id;
                                return id;
                            }
                        }

                        if (noCreate)
                        {
                            if (noException)
                                return -1;

                            isExpFinal = true;
                            throw new MetaStringsException($"Names.GetId cannot create new name: {name}", lastExp);
                        }
                    }
                    catch (Exception exp)
                    {
                        if (isExpFinal)
                            throw exp;

                        lastExp = exp;
                    }
                }

                throw new MetaStringsException("Names.GetId fails after a few tries", lastExp);
            }
            finally
            {
                ScopeTiming.RecordScope("Names.GetId", totalTimer);
            }
        }

        public static async Task<NameObj> GetNameAsync(Context ctxt, short id)
        {
            var totalTimer = ScopeTiming.StartTiming();
            try
            {
                if (id < 0)
                    return null;

                NameObj obj;
                if (sm_cacheBack.TryGetValue(id, out obj))
                    return obj;

                string sql = $"SELECT typeid, name, isNumeric FROM names WHERE id = {id}";
                using (var cmd = new MySqlCommand(sql, ctxt.DbConn.MyDb, ctxt.DbConn.MyTrans))
                using (var reader = await cmd.ExecuteReaderAsync().ConfigureAwait(false))
                {
                    if (!await reader.ReadAsync().ConfigureAwait(false))
                        throw new MetaStringsException($"Names.GetName fails to find record: {id}");

                    obj = new NameObj()
                    {
                        id = id,
                        typeId = reader.GetInt16(0),
                        name = reader.GetString(1),
                        isNumeric = reader.GetBoolean(2)
                    };
                    sm_cacheBack[id] = obj;
                    return obj;
                }
            }
            finally
            {
                ScopeTiming.RecordScope("Names.GetName", totalTimer);
            }
        }

        public static async Task CacheNamesAsync(Context ctxt, IEnumerable<short> ids)
        {
            var totalTimer = ScopeTiming.StartTiming();
            try
            {
                var stillToGet = ids.Where(id => !sm_cacheBack.ContainsKey(id));
                if (!stillToGet.Any())
                    return;

                var nameIdInPart = string.Join(",", stillToGet.Select(i => i.ToString()));
                var sql = $"SELECT id, typeid, name, isNumeric FROM names WHERE id IN ({nameIdInPart})";
                if (sql.Length > 64 * 1024)
                    throw new MetaStringsException("GetNames query exceeds MySQL batch limit of 64K.  Use a smaller batch of items.");

                using (var cmd = new MySqlCommand(sql, ctxt.DbConn.MyDb, ctxt.DbConn.MyTrans))
                using (var reader = await cmd.ExecuteReaderAsync().ConfigureAwait(false))
                {
                    while (await reader.ReadAsync().ConfigureAwait(false))
                    {
                        short id = reader.GetInt16(0);
                        sm_cacheBack[id] = 
                            new NameObj() 
                            { 
                                id = id, 
                                typeId = reader.GetInt16(1), 
                                name = reader.GetString(2),
                                isNumeric = reader.GetBoolean(3)
                            };
                    }
                }
            }
            finally
            {
                ScopeTiming.RecordScope("Names.CacheNames", totalTimer);
            }
        }

        public static void ClearCaches()
        {
            sm_cache.Clear();
            sm_cacheBack.Clear();
        }

        private static ConcurrentDictionary<string, short> sm_cache = new ConcurrentDictionary<string, short>();
        private static ConcurrentDictionary<short, NameObj> sm_cacheBack = new ConcurrentDictionary<short, NameObj>();
    }
}
