﻿using System;
using System.Threading;
using System.Threading.Tasks;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Net;

namespace metastrings
{ 
    public class Server : IDisposable
    {
        public Server(int port, bool log = false, bool dump = false, bool time = true)
        {
            m_log = log;
            m_dump = dump;

            Console.WriteLine($"metastrings server starting on port {port}: log: {log} - dump: {dump} - time: {time}");

            ScopeTiming.Init(time);

            m_listener = new HttpListener();
            m_listener.Prefixes.Add($"http://localhost:{port}/");

            Task.Run(async () => { await ServeAsync(); });
        }

        public void Dispose()
        {
            m_keepRunning = false;

            if (m_listener != null)
            {
                m_listener.Abort();
                m_listener = null;
            }

            while (m_inPlay > 0)
                Thread.Sleep(10);
        }

        private async Task ServeAsync()
        {
            m_listener.Start();

            if (m_log) Console.WriteLine("receiving requests");
            while (m_keepRunning)
            {
                try
                {
                    var ctxt = await m_listener.GetContextAsync();

                    // NOTE: Can't ConfigureAwait(false) in ASP.NET code
                    Task task = Task.Run(async () => await ProcessContextAsync(ctxt)); 
                }
                catch
                {
                    if (!m_keepRunning)
                        return;
                    else
                        throw;
                }
            }
        }

        private async Task ProcessContextAsync(HttpListenerContext httpCtxt)
        {
            var swTotal = ScopeTiming.StartTiming();
            try
            {
                var sw = ScopeTiming.StartTiming();
                Interlocked.Increment(ref m_inPlay);

                string type = httpCtxt.Request.QueryString["type"];
                if (m_log) Console.WriteLine("Request: {0} - {1} - {2}", httpCtxt.Request.RemoteEndPoint, httpCtxt.Request.Url, type);

                httpCtxt.Response.ContentEncoding = Encoding.UTF8;
                httpCtxt.Response.ContentType = "application/json";

                string requestBody;
                using (var sr = new StreamReader(httpCtxt.Request.InputStream, Encoding.UTF8))
                    requestBody = await sr.ReadToEndAsync(); // nope: .ConfigureAwait(false);
                ScopeTiming.RecordScope($"Request.{type}.Read", sw);
                if (m_dump) Console.WriteLine("Request Body - {0}:\r\n{1}\r\n", type, requestBody);
                else if (m_log) Console.WriteLine("Request - {0}: {1} long", type, requestBody.Length);

                string responseBody = "";
                using (var msCtxt = new Context())
                using (var msCmd = new Command(msCtxt))
                {
                    ScopeTiming.RecordScope($"Request.{type}.ReadyToProcess", sw);
                    switch (type)
                    {
                        case "get":
                            {
                                var getRequest = Utils.Deserialize<GetRequest>(requestBody);
                                var getResponses = await msCmd.GetAsync(getRequest);
                                responseBody = Utils.Serialize(getResponses);
                            }
                            break;

                        case "queryGet":
                            {
                                var getRequest = Utils.Deserialize<QueryGetRequest>(requestBody);
                                var getResponses = await msCmd.QueryGetAsync(getRequest);
                                responseBody = Utils.Serialize(getResponses);
                            }
                            break;

                        case "query":
                            {
                                var query = Utils.Deserialize<Select>(requestBody);

                                StringBuilder sb = new StringBuilder();
                                using (var reader = await msCmd.GetReaderAsync(query))
                                using (var writer = new StreamWriter(httpCtxt.Response.OutputStream, Encoding.UTF8))
                                {
                                    sb.Append(CsvPrep.CreateCsv(reader.FieldNames));
                                    sb.Append('\n');

                                    List<string> rowStrs = new List<string>(reader.FieldCount);
                                    while (await reader.ReadAsync())
                                    {
                                        reader.GetCsv(rowStrs);
                                        sb.Append(CsvPrep.CreateCsv(rowStrs));
                                        sb.Append('\n');
                                        rowStrs.Clear();

                                        if (sb.Length > 16 * 1024)
                                        {
                                            await writer.WriteAsync(sb.ToString());
                                            sb.Clear();
                                        }
                                    }

                                    await writer.WriteAsync(sb.ToString());
                                    sb.Clear();
                                }
                            }
                            break;

                        case "define":
                            {
                                var define = Utils.Deserialize<Define>(requestBody);
                                await msCmd.DefineAsync(define);
                            }
                            break;

                        case "definecsv":
                            {
                                var csv = requestBody;
                                await msCmd.DefineCsvAsync(csv);
                            }
                            break;

                        case "delete":
                            {
                                var del = Utils.Deserialize<Delete>(requestBody);
                                await msCmd.DeleteAsync(del);
                            }
                            break;

                        case "sql":
                            {
                                Select query = Utils.Deserialize<Select>(requestBody);
                                responseBody = await msCmd.GenerateSqlAsync(query);
                                httpCtxt.Response.ContentType = "text/plain";
                            }
                            break;

                        case "timing":
                            {
                                var timing = Utils.Deserialize<Timing>(requestBody);
                                string timingSummary = ScopeTiming.Summary;
                                if (timing.reset)
                                    ScopeTiming.Clear();
                                responseBody = timingSummary;
                                httpCtxt.Response.ContentType = "text/plain";
                            }
                            break;

                        case "drop":
                            {
                                var drop = Utils.Deserialize<Drop>(requestBody);
                                await msCmd.DropAsync(drop);
                            }
                            break;

                        case "reset":
                            {
                                var reset = Utils.Deserialize<Reset>(requestBody);
                                msCmd.Reset(reset);
                            }
                            break;

                        default:
                            throw new MetaStringsException($"Invalid query: {type}");
                    }

                    await msCtxt.ProcessPostOpsAsync();
                    
                    ScopeTiming.RecordScope($"Request.{type}.Process", sw);
                }

                if (m_dump) Console.WriteLine("Response {0} Body:\r\n{1}\r\n", type, responseBody);
                else if (m_log) Console.WriteLine("Response {0}: {1} long", type, responseBody.Length);

                if (responseBody != "")
                {
                    using (StreamWriter outputWriter = new StreamWriter(httpCtxt.Response.OutputStream, Encoding.UTF8))
                        await outputWriter.WriteAsync(responseBody);
                }
                else
                    httpCtxt.Response.OutputStream.Close();

                ScopeTiming.RecordScope($"Request.{type}.Respond", sw);
            }
            catch (Exception exp)
            {
                Console.WriteLine("Server ERROR:\n{0}", exp);

                while (exp.InnerException != null)
                    exp = exp.InnerException;

                httpCtxt.Response.StatusCode = 500;
                httpCtxt.Response.OutputStream.Write(Encoding.UTF8.GetBytes($"{exp.GetType().FullName}: {exp.Message}"));
                httpCtxt.Response.OutputStream.Close();
            }
            finally
            {
                Interlocked.Decrement(ref m_inPlay);
                ScopeTiming.RecordScope("Request.Total", swTotal);
            }
        }

        private bool m_log = true;
        private bool m_dump = true;

        private bool m_keepRunning = true;
        
        private HttpListener m_listener;

        private int m_inPlay = 0;
    }
}
