﻿using System;
using System.Threading.Tasks;
using System.Collections.Generic;
using System.Text;
using System.Linq;

using MySql.Data.MySqlClient;

namespace metastrings
{
    public static class Sql
    {
        public static async Task<string> GenerateSqlAsync(Context ctxt, Select query)
        {
            //
            // "COMPILE"
            //
            if (string.IsNullOrWhiteSpace(query.type))
                throw new MetaStringsException("Invalid query, TypeName is missing");

            if (query.select == null || query.select.Count == 0)
                throw new MetaStringsException("Invalid query, SelectColumns is empty");

            if (query.orderBy != null)
            {
                foreach (var order in query.orderBy)
                {
                    string orderField = order.field.Trim();
                    if (!query.select.Contains(orderField))
                        throw new MetaStringsException($"Invalid query, OrderByColumns must be in SelectColumns: {order.field.Trim()}");
                }
            }

            if (query.where != null)
            {
                foreach (var criteriaSet in query.where)
                {
                    foreach (var criteria in criteriaSet.criteria)
                    {
                        ValidateOperator(criteria.op);
                        ValidateParameterName(criteria.paramName);
                    }
                }
            }


            //
            // SETUP
            //
            short typeId = await Types.GetIdAsync(ctxt, query.type, noCreate: true).ConfigureAwait(false);
            TypeObj typeObj = await Types.GetTypeAsync(ctxt, typeId).ConfigureAwait(false);

            // Gather columns
            var names = new List<string>();

            names.AddRange(query.select);

            if (query.orderBy != null)
                names.AddRange(query.orderBy.Select(o => o.field));

            if (query.where != null)
            {
                foreach (var criteriaSet in query.where)
                    names.AddRange(criteriaSet.criteria.Select(w => w.name));
            }

            // Cut them down
            names = names.Select(n => n.Trim()).Where(n => !string.IsNullOrWhiteSpace(n)).Distinct().ToList();

            // Get name objects
            var nameObjs = new Dictionary<string, NameObj>(names.Count);
            foreach (var name in names)
            {
                if (name == "value" || name == "id")
                {
                    nameObjs.Add(name, null);
                }
                else
                {
                    var nameId = await Names.GetIdAsync(ctxt, typeId, name, noCreate: true).ConfigureAwait(false);
                    var nameObj = await Names.GetNameAsync(ctxt, nameId);
                    nameObjs.Add(name, nameObj);
                }
            }


            //
            // SELECT
            //
            string selectPart = "";
            foreach (var name in query.select.Select(n => n.Trim()).Where(n => !string.IsNullOrWhiteSpace(n)))
            {
                var cleanName = CleanName(name);

                if (selectPart.Length > 0)
                    selectPart += ",\r\n";

                if (name == "value")
                {
                    if (typeObj.isNumeric)
                        selectPart += "bv.numberValue";
                    else
                        selectPart += query.prefixesOnly ? "bv.prefix" : "IFNULL(bv.stringValue, bv.prefix)";
                }
                else if (name == "id")
                    selectPart += "i.id";
                else if (name == "count")
                    selectPart += "COUNT(*)";
                else if (nameObjs[name] == null)
                    selectPart += "NULL";
                else
                    selectPart += $"(CASE WHEN iv{cleanName}.isNumeric THEN iv{cleanName}.numberValue ELSE IFNULL(iv{cleanName}.stringValue, iv{cleanName}.prefix) END)";

                selectPart += $" AS {cleanName}";
            }
            selectPart = "SELECT\r\n" + selectPart;


            //
            // FROM
            //
            string fromPart = "FROM\r\nitems AS i";
            if (nameObjs.ContainsKey("value"))
                fromPart += "\r\nJOIN bvalues bv ON bv.id = i.valueid";

            foreach (var name in names.Select(n => n.Trim()).Where(n => !string.IsNullOrWhiteSpace(n)))
            {
                if 
                (
                    name != "value" 
                    && 
                    name != "id"
                    && 
                    nameObjs.ContainsKey(name) 
                    && 
                    nameObjs[name] != null
                )
                {
                    var cleanName = CleanName(name);
                    fromPart += 
                        $"\r\nLEFT OUTER JOIN itemvalues as iv{cleanName} ON iv{cleanName}.itemid = i.id" +
                        $" AND iv{cleanName}.nameid = {nameObjs[name].id}";
                }
            }


            //
            // WHERE
            //
            string wherePart = $"i.typeid = {typeId}";
            if (query.where != null)
            {
                foreach (var criteriaSet in query.where)
                {
                    if (criteriaSet.criteria.Count == 0)
                        continue;

                    wherePart += "\r\nAND\r\n";

                    wherePart += "(";
                    bool addedOneYet = false;
                    foreach (var where in criteriaSet.criteria)
                    {
                        string name = where.name.Trim();
                        if (string.IsNullOrWhiteSpace(name))
                            continue;

                        if (!addedOneYet)
                            addedOneYet = true;
                        else
                            wherePart += $" {Enum.GetName(criteriaSet.combine.GetType(), criteriaSet.combine)} ";

                        var nameObj = nameObjs[name];
                        var cleanName = CleanName(name);

                        if (nameObj == null && cleanName != "value") // something on the item row
                        {
                            wherePart += $"i.{cleanName} {where.op} {where.paramName}";
                        }
                        else if (where.op.Equals("matches", StringComparison.OrdinalIgnoreCase))
                        {
                            string columnPart = 
                                name == "value" 
                                ? "bv.prefix, bv.stringValue" 
                                : $"iv{cleanName}.prefix, iv{cleanName}.stringValue";
                            wherePart += $"MATCH({columnPart}) AGAINST ({where.paramName} IN BOOLEAN MODE) > 0";
                            if (query.orderBy == null || query.orderBy.Count == 0)
                            {
                                selectPart += $", MATCH({columnPart}) AGAINST ({where.paramName} IN BOOLEAN MODE) AS {cleanName}Score";
                                query.orderBy = new List<Order> { new Order() { field = $"{cleanName}Score", descending = true } };
                            }
                        }
                        else if (cleanName == "value")
                        {
                            if (typeObj.isNumeric)
                                wherePart += $"bv.numberValue {where.op} {where.paramName}";
                            else
                                wherePart += $"bv.prefix {where.op} {where.paramName}";
                        }
                        else if (nameObj.isNumeric)
                        {
                            wherePart += $"iv{cleanName}.numberValue {where.op} {where.paramName}";
                        }
                        else
                        {
                            wherePart += $"iv{cleanName}.prefix {where.op} {where.paramName}";
                        }
                    }
                    wherePart += ")";
                }
            }
            wherePart = "WHERE " + wherePart;


            //
            // ORDER BY
            //
            string orderBy = "";
            if (query.orderBy != null)
            {
                foreach (var order in query.orderBy)
                {
                    if (string.IsNullOrWhiteSpace(order.field))
                        continue;

                    if (orderBy.Length > 0)
                        orderBy += ",\r\n";

                    string orderColumn = CleanName(order.field);

                    orderBy += MySqlHelper.EscapeString(orderColumn) + (order.descending ? " DESC" : " ASC");
                }

                if (orderBy.Length > 0)
                    orderBy = "ORDER BY\r\n" + orderBy;
            }


            //
            // LIMIT
            //
            string limitPart = "";
            if (query.limit > 0)
                limitPart = $"LIMIT\r\n{query.limit}";


            //
            // SQL
            //
            StringBuilder sb = new StringBuilder();

            sb.Append($"{selectPart.Trim()}");
            
            sb.Append($"\r\n\r\n{fromPart}");

            if (!string.IsNullOrWhiteSpace(wherePart))
            {
                sb.Append($"\r\n\r\n{wherePart}");
            }

            if (!string.IsNullOrWhiteSpace(orderBy))
            {
                sb.Append($"\r\n\r\n{orderBy}");
            }

            if (!string.IsNullOrWhiteSpace(limitPart))
            {
                sb.Append($"\r\n\r\n{limitPart}");
            }

            string sql = sb.ToString();
            return sql;
        }

        private static void ValidateOperator(string op)
        {
            if (!sm_validOps.Contains(op.ToLower()))
                throw new MetaStringsException($"Invalid query operator: {op}");
        }

        private static void ValidateParameterName(string parm)
        {
            if (string.IsNullOrWhiteSpace(parm))
                throw new MetaStringsException($"Invalid parameter name: (blank)");

            parm = parm.Trim();

            if (parm.Length < 2 || parm[0] != '@')
                throw new MetaStringsException($"Invalid parameter name: {parm}");

            for (int p = 1; p < parm.Length; ++p)
            {
                if (!char.IsLetterOrDigit(parm[p]))
                    throw new MetaStringsException($"Invalid parameter name: {parm}");
            }
        }

        public static string CleanName(string name) // used for table and column aliases
        {
            string clean = "";
            foreach (char c in name)
            {
                if (char.IsLetterOrDigit(c))
                {
                    clean += c;
                }
            }

            if (string.IsNullOrWhiteSpace(clean) || !char.IsLetter(clean[0]))
            {
                clean = "a" + clean;
            }
            return clean;
        }

        private static HashSet<string> sm_validOps =
            new HashSet<string>
            (
                new[]
                {
                    "=", "<>", ">", ">=", "<", "<=",
                    "matches",
                    "like"
                }
            );
    }
}
