﻿using System;
using System.Collections.Concurrent;
using System.Threading.Tasks;

using MySql.Data.MySqlClient;

namespace metastrings
{
    public class TypeObj
    {
        public short id;
        public string name;
        public bool isNumeric;
    }

    public static class Types
    {
        public static void Reset(Context ctxt)
        {
            sm_cache.Clear();
            sm_cacheBack.Clear();

            using (var cmd = new MySqlCommand("DELETE FROM types", ctxt.DbConn.MyDb, ctxt.DbConn.MyTrans))
                cmd.ExecuteNonQuery();
        }

        public static async Task<short> GetIdAsync(Context ctxt, string name, bool isNumeric = false, bool noCreate = false, bool noException = false)
        {
            var totalTimer = ScopeTiming.StartTiming();
            try
            {
                short id;
                if (sm_cache.TryGetValue(name, out id))
                    return id;

                Exception lastExp = null;
                bool isExpFinal = false;
                for (int tryCount = 1; tryCount <= 4; ++tryCount)
                {
                    try
                    {
                        string selectSql = "SELECT id FROM types WHERE name = @name";
                        using (var cmd = new MySqlCommand(selectSql, ctxt.DbConn.MyDb, ctxt.DbConn.MyTrans))
                        {
                            cmd.Parameters.AddWithValue("@name", name);
                            object idObj = await cmd.ExecuteScalarAsync().ConfigureAwait(false);
                            if (idObj != null && idObj != DBNull.Value)
                            {
                                id = Convert.ToInt16(idObj);
                                sm_cache[name] = id;
                                return id;
                            }
                        }

                        if (noCreate)
                        {
                            if (noException)
                                return -1;

                            isExpFinal = true;
                            throw new MetaStringsException($"Types.GetId cannot create new type: {name}", lastExp);
                        }

                        string insertSql = "INSERT types (name, isNumeric) VALUES (@name, @isNumeric)";
                        using (var cmd = new MySqlCommand(insertSql, ctxt.DbConn.MyDb, ctxt.DbConn.MyTrans))
                        {
                            cmd.Parameters.AddWithValue("@name", name);
                            cmd.Parameters.AddWithValue("@isNumeric", isNumeric);

                            await cmd.ExecuteNonQueryAsync().ConfigureAwait(false);

                            id = (short)cmd.LastInsertedId;
                            sm_cache[name] = id;

                            return id;
                        }
                    }
                    catch (Exception exp)
                    {
                        if (isExpFinal)
                            throw exp;

                        lastExp = exp;
                    }
                }

                throw new MetaStringsException("Types.GetId fails after a few tries", lastExp);
            }
            finally
            {
                ScopeTiming.RecordScope("Types.GetId", totalTimer);
            }
        }

        public static async Task<TypeObj> GetTypeAsync(Context ctxt, short id)
        {
            var totalTimer = ScopeTiming.StartTiming();
            try
            {
                if (id < 0)
                    return null;

                TypeObj obj;
                if (sm_cacheBack.TryGetValue(id, out obj))
                    return obj;

                string sql = $"SELECT name, isNumeric FROM types WHERE id = {id}";
                using (var cmd = new MySqlCommand(sql, ctxt.DbConn.MyDb, ctxt.DbConn.MyTrans))
                using (var reader = await cmd.ExecuteReaderAsync().ConfigureAwait(false))
                {
                    if (!await reader.ReadAsync().ConfigureAwait(false))
                        throw new MetaStringsException($"Types.GetType fails to find record: {id}");

                    obj =
                        new TypeObj()
                        {
                            id = id,
                            name = reader.GetString(0),
                            isNumeric = reader.GetBoolean(1)
                        };
                    sm_cacheBack[id] = obj;
                    return obj;
                }
            }
            finally
            {
                ScopeTiming.RecordScope("Types.GetType", totalTimer);
            }
        }

        public static void ClearCaches()
        {
            sm_cache.Clear();
            sm_cacheBack.Clear();
        }

        private static ConcurrentDictionary<string, short> sm_cache = new ConcurrentDictionary<string, short>();
        private static ConcurrentDictionary<short, TypeObj> sm_cacheBack = new ConcurrentDictionary<short, TypeObj>();
    }
}
