﻿using System;
using System.Collections.Generic;
using System.Collections.Concurrent;
using System.Threading;
using System.Threading.Tasks;
using System.Linq;
using System.Data;

using MySql.Data.MySqlClient;

namespace metastrings
{
    public static class Values
    {
        public const int PrefixLen = 255;

        public static void Reset(Context ctxt)
        {
            sm_cache.Clear();
            sm_cacheBack.Clear();

            using (var cmd = new MySqlCommand("DELETE FROM bvalues", ctxt.DbConn.MyDb, ctxt.DbConn.MyTrans))
                cmd.ExecuteNonQuery();
        }

        public static async Task<long> GetIdAsync(Context ctxt, object value)
        {
            var totalTimer = ScopeTiming.StartTiming();
            try
            {
                var localTimer = ScopeTiming.StartTiming();

                if (value == null)
                    return -1;

                long idFromCache;
                if (sm_cache.TryGetValue(value, out idFromCache))
                    return idFromCache;

                Exception lastExp = null;
                for (int tryCount = 1; tryCount <= 2; ++tryCount)
                {
                    if (value is string)
                    {
                        sm_writeStringLock.EnterWriteLock();
                        string strValue = (string)value;
                        try
                        {
                            string procText, lengthTxt;
                            if (strValue.Length < PrefixLen)
                            {
                                procText = "GetShortStringId";
                                lengthTxt = "short";
                            }
                            else
                            {
                                procText = "GetLongStringId";
                                lengthTxt = "long";
                            }

                            using (var cmd = new MySqlCommand(procText, ctxt.DbConn.MyDb, ctxt.DbConn.MyTrans))
                            {
                                cmd.CommandType = CommandType.StoredProcedure;
                                cmd.Parameters.AddWithValue("stringValueParam", strValue);
                                object idObj = await cmd.ExecuteScalarAsync().ConfigureAwait(false);
                                ScopeTiming.RecordScope($"Values.GetId.SELECT(string:{lengthTxt})", localTimer);
                                if (!(idObj == null || idObj == DBNull.Value))
                                {
                                    long id = Convert.ToInt64(idObj);
                                    sm_cache[value] = id;
                                    return id;
                                }
                            }
                        }
                        catch (Exception exp)
                        {
                            lastExp = exp;
                        }
                        finally
                        {
                            sm_writeStringLock.ExitWriteLock();
                        }
                    }
                    else 
                    {
                        sm_writeNumberLock.EnterWriteLock();
                        double numberValue = Convert.ToDouble(value);
                        try
                        {
                            using (var cmd = new MySqlCommand("GetNumberId", ctxt.DbConn.MyDb, ctxt.DbConn.MyTrans))
                            {
                                cmd.CommandType = CommandType.StoredProcedure;
                                cmd.Parameters.AddWithValue("numberValueParam", numberValue);
                                object idObj = await cmd.ExecuteScalarAsync().ConfigureAwait(false);
                                ScopeTiming.RecordScope("Values.GetId.SELECT(number)", localTimer);
                                if (!(idObj == null || idObj == DBNull.Value))
                                {
                                    long id = Convert.ToInt64(idObj);
                                    sm_cache[value] = id;
                                    return id;
                                }
                            }
                        }
                        catch (Exception exp)
                        {
                            lastExp = exp;
                        }
                        finally
                        {
                            sm_writeNumberLock.ExitWriteLock();
                        }
                    }
                }

                throw new MetaStringsException("Values.GetId failed after a few retries", lastExp);
            }
            finally
            {
                ScopeTiming.RecordScope("Values.GetId", totalTimer);
            }
        }

        public static async Task<object> GetValueAsync(Context ctxt, long id)
        {
            var totalTimer = ScopeTiming.StartTiming();
            try
            {
                object objFromCache;
                if (sm_cacheBack.TryGetValue(id, out objFromCache))
                    return objFromCache;

                string sql = $"SELECT isNumeric, numberValue, IFNULL(stringValue, prefix) FROM bvalues WHERE id = {id}";
                using (var cmd = new MySqlCommand(sql, ctxt.DbConn.MyDb, ctxt.DbConn.MyTrans))
                using (var reader = await cmd.ExecuteReaderAsync().ConfigureAwait(false))
                { 
                    if (!await reader.ReadAsync().ConfigureAwait(false))
                        throw new MetaStringsException("Values.GetValue fails to find record with ID = " + id);

                    object toReturn;
                    bool isNumeric = reader.GetBoolean(0);
                    if (isNumeric)
                        toReturn = reader.GetDouble(1);
                    else
                        toReturn = reader.GetString(2);
                    sm_cacheBack[id] = toReturn;
                    return toReturn;
                }
            }
            finally
            {
                ScopeTiming.RecordScope("Values.GetValue", totalTimer);
            }
        }

        public static async Task CacheValuesAsync(Context ctxt, IEnumerable<long> ids)
        {
            var totalTimer = ScopeTiming.StartTiming();
            try
            {
                var stillToGet = ids.Where(id => !sm_cacheBack.ContainsKey(id));
                if (!stillToGet.Any())
                    return;

                var valueIdInPart = string.Join(",", stillToGet.Select(i => i.ToString()));
                var sql = $"SELECT id, isNumeric, numberValue, IFNULL(stringValue, prefix) FROM bvalues WHERE id IN ({valueIdInPart})";
                if (sql.Length > 64 * 1024)
                    throw new MetaStringsException("GetValues query exceeds MySQL batch limit of 64K.  Use a smaller batch of items.");

                using (var cmd = new MySqlCommand(sql, ctxt.DbConn.MyDb, ctxt.DbConn.MyTrans))
                using (var reader = await cmd.ExecuteReaderAsync().ConfigureAwait(false))
                {
                    while (await reader.ReadAsync().ConfigureAwait(false))
                    {
                        long id = reader.GetInt64(0);

                        bool isNumeric = reader.GetBoolean(1);

                        object obj;
                        if (isNumeric)
                            obj = reader.GetDouble(2);
                        else
                            obj = reader.GetString(3);

                        sm_cacheBack[id] = obj;
                    }
                }
            }
            finally
            {
                ScopeTiming.RecordScope("Values.CacheValues", totalTimer);
            }
        }

        public static void ClearCaches()
        {
            sm_cache.Clear();
            sm_cacheBack.Clear();
        }

        // Prevent concurrent INSERTs as there's no enforced primary keep on the bvalues table
        private static ReaderWriterLockSlim sm_writeStringLock = new ReaderWriterLockSlim();
        private static ReaderWriterLockSlim sm_writeNumberLock = new ReaderWriterLockSlim();

        private static ConcurrentDictionary<object, long> sm_cache = new ConcurrentDictionary<object, long>();
        private static ConcurrentDictionary<long, object> sm_cacheBack = new ConcurrentDictionary<long, object>();
    }
}
