﻿using System;
using System.Threading.Tasks;
using System.Collections.Generic;
using System.Configuration;
using System.Diagnostics;
using System.Text.RegularExpressions;
using System.IO;

using MySql.Data.MySqlClient;

namespace metastrings
{
    class Program
    {
        static void Main(string[] args)
        {
            if (args.Length == 0)
            {
                Console.WriteLine("Usage: metafiles <path to directory to index and query> ...");
                return;
            }

            var dirPaths = args;
            foreach (var dirPath in dirPaths)
            {
                if (!Directory.Exists(dirPath))
                {
                    Console.WriteLine("ERROR: Directory does not exist: {0}", dirPath);
                    return;
                }
            }

            var fileInfo = MyFileInfoMgr.GetFileInfo(dirPaths);
            Console.WriteLine("Ready with {0} files", fileInfo.Count);

            // Create server for our API to talk to...all in this one process
            string serverUrl = "http://localhost:9914";
            using (var metastringsServer = new Server(9914))
            using (var metastringsApi = new api(serverUrl))
            using (var metafilesDb = new MySqlConnection(ConfigurationManager.ConnectionStrings["metafiles"].ConnectionString))
            {
                metafilesDb.Open();

                for (int run = 1; run <= 2; ++run)
                {

                    Console.WriteLine();
                    Console.WriteLine("//");
                    Console.WriteLine("// RUN # {0} !!!", run);
                    Console.WriteLine("//");
                    Console.WriteLine();

                    Console.Write("Resetting metafiles...");
                    using (var cmd = new MySqlCommand("TRUNCATE TABLE fi", metafilesDb))
                        cmd.ExecuteNonQuery();
                    Console.WriteLine(" Done!");
                    Console.WriteLine();

                    if (run == 1)
                    {
                        Console.Write("Resetting metastrings (deep)...");
                        metastringsApi.ResetAsync(new Reset() { includeNameValues = true }).Wait(); // frozen
                        Console.WriteLine(" Done!");
                        Console.WriteLine();
                    }
                    else if (run == 2)
                    {
                        Console.Write("Resetting metastrings (shallow)...");
                        metastringsApi.ResetAsync(new Reset()).Wait(); // cold
                        Console.WriteLine(" Done!");
                        Console.WriteLine();
                    }

                    {
                        Console.WriteLine("metastrings...");
                        Stopwatch sw = Stopwatch.StartNew();

                        // Divide and conquer!
                        int metastringsParallelism = Environment.ProcessorCount * 8;
                        var fileBuckets = Utils.SplitUp(fileInfo.Keys, metastringsParallelism);
                        List<Task> tasks = new List<Task>();
                        foreach (var bucket in fileBuckets)
                        {
                            tasks.Add
                            (
                                Task.Run(async () =>
                                await ProcessMetaStringsBucketAsync
                                (
                                    serverUrl,
                                    bucket,
                                    fileInfo
                                ).ConfigureAwait(false))
                            );
                        }
                        Task.WaitAll(tasks.ToArray());

                        Console.WriteLine("metastrings update elapsed: {0} ms", sw.ElapsedMilliseconds);
                        sw.Restart();
                    }

                    {
                        Console.WriteLine();
                        Console.WriteLine("metafiles...");
                        Stopwatch sw = Stopwatch.StartNew();
                        using (var fqlTrans = metafilesDb.BeginTransaction())
                        {
                            var sql =
                                "INSERT fi (pathPrefix, path, length, created, lastModified, ext, readOnly, namePrefix, name, company, product, descriptionPrefix, description, version, language, copyright)"
                                + " VALUES (@pathPrefix, @path, @length, @created, @lastModified, @ext, @readOnly, @namePrefix, @name, @company, @product, @descriptionPrefix, @description, @version, @language, @copyright)"
                                + " ON DUPLICATE KEY UPDATE ext = @ext"; // something
                            foreach (var newResult in fileInfo.Keys)
                            {
                                var newData = fileInfo[newResult].Metadata;
                                using (var cmd = new MySqlCommand(sql, metafilesDb))
                                {
                                    cmd.Parameters.AddWithValue("@pathPrefix", newResult.Substring(0, Math.Min(newResult.Length, 255)));
                                    cmd.Parameters.AddWithValue("@path", newResult);
                                    cmd.Parameters.AddWithValue("@length", newData["length#"]);
                                    cmd.Parameters.AddWithValue("@created", newData["created"]);
                                    cmd.Parameters.AddWithValue("@lastModified", newData["lastModified"]);
                                    cmd.Parameters.AddWithValue("@ext", newData["ext"]);
                                    cmd.Parameters.AddWithValue("@readOnly", newData["readOnly"]);
                                    string name = (string)newData["name"];
                                    cmd.Parameters.AddWithValue("@namePrefix", name.Substring(0, Math.Min(name.Length, 255)));
                                    cmd.Parameters.AddWithValue("@name", name);
                                    cmd.Parameters.AddWithValue("@company", newData["company"]);
                                    cmd.Parameters.AddWithValue("@product", newData["product"]);
                                    string description = (string)newData["description"];
                                    cmd.Parameters.AddWithValue("@descriptionPrefix", description.Substring(0, Math.Min(description.Length, 255)));
                                    cmd.Parameters.AddWithValue("@description", newData["description"]);
                                    cmd.Parameters.AddWithValue("@version", newData["version"]);
                                    cmd.Parameters.AddWithValue("@language", newData["language"]);
                                    cmd.Parameters.AddWithValue("@copyright", newData["copyright"]);
                                    cmd.ExecuteNonQuery();
                                }
                            }
                            fqlTrans.Commit();
                        }

                        Console.WriteLine("metafiles added elapsed: {0} ms", sw.ElapsedMilliseconds);
                        sw.Restart();
                    }

                    Console.WriteLine();
                    var summary = metastringsApi.GetTimingSummary(new Timing() { reset = true }).Result;
                    Console.WriteLine(summary);
                }

                while (true)
                {
                    Console.WriteLine();
                    Console.Write("> ");
                    string line = Console.ReadLine();
#if !DEBUG
                    try
#endif
                    {
                        var regex = new Regex(@"^(\S+) (\S+) (\S+)$");
                        var match = regex.Match(line);
                        if (!match.Success || match.Groups.Count != 4)
                        {
                            Console.WriteLine("Sorry, invalid query: has to look like... ext = exe");
                            continue;
                        }

                        string queryName = match.Groups[1].Value;
                        string op = match.Groups[2].Value;
                        string on = match.Groups[3].Value;

                        Console.WriteLine("Field: {0}", queryName);
                        Console.WriteLine("Op:    {0}", op);
                        Console.WriteLine("Value: {0}", on);

                        {
                            Console.WriteLine("metastrings...");
                            Stopwatch sw = Stopwatch.StartNew();
                            Select select = new Select();
                            select.select = new List<string>() { "value" };
                            select.type = "fileInfo";
                            select.where = CriteriaSet.GenWhere(new Criteria() { name = queryName == "path" ? "value" : queryName, op = op, paramName = "@on" });
                            select.cmdParams = new Dictionary<string, object>();
                            select.cmdParams["@on"] = on;
                            int resultCount = 0;
                            int printed = 0;
                            using (var query = new query(serverUrl))
                            {
                                query.ExecuteAsync(select).Wait();
                                while (query.ReadAsync().Result)
                                {
                                    if (printed < 10)
                                    {
                                        Console.WriteLine("{0}", query[0]);
                                        ++printed;
                                    }
                                    else
                                        Console.Write('.');
                                    ++resultCount;
                                }
                            }
                            Console.WriteLine();
                            Console.WriteLine("metastrings {0} results, elapsed: {1} ms", resultCount, sw.ElapsedMilliseconds);
                        }

                        {
                            Console.WriteLine("metafiles...");
                            Stopwatch sw = Stopwatch.StartNew();
                            string sql = $"SELECT path FROM fi WHERE ";
                            if (op == "matches")
                                sql += $"MATCH({queryName}) AGAINST (@on IN BOOLEAN MODE) > 0";
                            else
                                sql += $"{queryName} {op} @on";
                            int resultCount = 0;
                            int printed = 0;
                            using (var cmd = new MySqlCommand(sql, metafilesDb))
                            {
                                cmd.Parameters.AddWithValue("@on", on);
                                using (var reader = cmd.ExecuteReader())
                                {
                                    while (reader.Read())
                                    {
                                        if (printed < 10)
                                        {
                                            Console.WriteLine("{0}", reader[0]);
                                            ++printed;
                                        }
                                        else
                                            Console.Write('.');
                                        ++resultCount;
                                    }
                                }
                            }
                            Console.WriteLine();
                            Console.WriteLine("metafiles {0} results, elased: {1} ms", resultCount, sw.ElapsedMilliseconds);
                        }
                    }
#if !DEBUG
                    catch (Exception exp)
                    {
                        while (exp.InnerException != null)
                            exp = exp.InnerException;

                        Console.WriteLine("ERROR:\r\n{0}: {1}", exp.GetType().FullName, exp.Message);
                    }
#endif
                }
            }
        }

        static async Task ProcessMetaStringsBucketAsync(string serverUrl, List<string> filePaths, Dictionary<string, MyFileInfo> fileInfos)
        {
            var curFileInfos = new List<MyFileInfo>();
            const int maxFilesPerBatch = 100;
            foreach (string filePath in filePaths)
            {
                curFileInfos.Add(fileInfos[filePath]);
                if (curFileInfos.Count >= maxFilesPerBatch)
                {
                    await ProcessCsvAsync(serverUrl, curFileInfos).ConfigureAwait(false);
                    curFileInfos.Clear();
                }
            }

            if (curFileInfos.Count > 0)
                await ProcessCsvAsync(serverUrl, curFileInfos).ConfigureAwait(false);
        }

        static async Task ProcessCsvAsync(string serverUrl, List<MyFileInfo> fileInfos)
        {
            var headersDict = new ListDictionary<string, string>();
            headersDict.Set("type", "fileInfo");

            var namesList = new List<string>(fileInfos[0].Metadata.Keys);

            using (var define = new define(serverUrl))
            {
                await define.BeginAsync(headersDict, namesList).ConfigureAwait(false);
                List<object> valuesList = new List<object>(namesList.Count);
                foreach (var fileInfo in fileInfos)
                {
                    valuesList.Clear();

                    valuesList.Add(fileInfo.FilePath);
                    foreach (var name in namesList)
                    {
                        object obj = fileInfo.Metadata.ContainsKey(name) ? fileInfo.Metadata[name] : null;
                        valuesList.Add(obj);
                    }

                    await define.WriteRecordAsync(valuesList).ConfigureAwait(false);
                }
                await define.EndAsync().ConfigureAwait(false);
            }
        }
    }
}
