﻿using System;
using System.Collections.Generic;
using System.IO;

using Microsoft.VisualStudio.TestTools.UnitTesting;

namespace metastrings
{
    [TestClass]
    public class ApiServerTests
    {
        [TestMethod]
        public void TestApiServer()
        {
            string serverUrl = "http://localhost:9914";
            using (var server = new Server(9914))
            using (var api = new api(serverUrl))
            {
                // Create two items with metadata
                DateTime definedDate = DateTime.UtcNow;
                {
                    var define1 = new Define() { type = "ApiTable" };
                    define1.SetData("Key1", "Name1", "Value1");
                    define1.SetData("Key1", "Name2", "Value2");
                    api.DefineAsync(define1).Wait();

                    var define2 = new Define() { type = "ApiTable" };
                    define2.SetData("Key2", "Name3", "Value3");
                    define2.SetData("Key2", "Name4", "Value4");
                    api.DefineAsync(define2).Wait();
                }

                // Verify that we can get what we just defined
                {
                    var gotten = api.GetAsync(new GetRequest() { type = "ApiTable", values = new List<object> { "Key1" } }).Result;
                    Assert.AreEqual(1, gotten.metadata.Count);
                    var response = gotten.metadata[0];
                    Assert.AreEqual(2, response.Count);
                    Assert.AreEqual("Value1", response["Name1"]);
                    Assert.AreEqual("Value2", response["Name2"]);
                }

                {
                    var gotten = api.GetAsync(new GetRequest() { type = "ApiTable", values = new List<object> { "Key2" } }).Result;
                    Assert.AreEqual(1, gotten.metadata.Count);
                    var response = gotten.metadata[0];
                    Assert.AreEqual(2, response.Count);
                    Assert.AreEqual("Value3", response["Name3"]);
                    Assert.AreEqual("Value4", response["Name4"]);
                }

                {
                    var sqlObj = new QueryGetRequest();
                    sqlObj.type = "ApiTable";
                    sqlObj.where = CriteriaSet.GenWhere(new Criteria() { name = "Name1", op = "=", paramName = "@name1" });
                    sqlObj.orderBy = new List<Order> { new Order() { field = "value", descending = false } };
                    sqlObj.cmdParams = new Dictionary<string, object>();
                    sqlObj.cmdParams["@name1"] = "Value1";

                    var gotten = api.QueryGetAsync(sqlObj).Result;
                    var response = gotten.metadata;
                    Assert.AreEqual(1, response.Count);
                    Assert.AreEqual(3, response[0].Count);
                    Assert.AreEqual("Key1", response[0]["value"]);
                    Assert.AreEqual("Value1", response[0]["Name1"]);
                    Assert.AreEqual("Value1", response[0]["Name1"]);
                }

                // prefixes-only
                {
                    var sqlObj = new QueryGetRequest();
                    sqlObj.type = "ApiTable";
                    sqlObj.where = CriteriaSet.GenWhere(new Criteria() { name = "Name1", op = "=", paramName = "@name1" });
                    sqlObj.orderBy = new List<Order> { new Order() { field = "value", descending = false } };
                    sqlObj.cmdParams = new Dictionary<string, object>();
                    sqlObj.cmdParams["@name1"] = "Value1";
                    sqlObj.prefixesOnly = true;

                    var gotten = api.QueryGetAsync(sqlObj).Result;
                    var response = gotten.metadata;
                    Assert.AreEqual(1, response.Count);
                    Assert.AreEqual(3, response[0].Count);
                    Assert.AreEqual("Key1", response[0]["value"]);
                    Assert.AreEqual("Value1", response[0]["Name1"]);
                    Assert.AreEqual("Value1", response[0]["Name1"]);
                }

                {
                    var sqlObj = new Select();
                    sqlObj.select = new List<string> { "value", "Name1", "Name2", "Name3", "Name4" };
                    sqlObj.type = "ApiTable";
                    sqlObj.where = CriteriaSet.GenWhere(new Criteria() { name = "Name1", op = "=", paramName = "@name1" });
                    sqlObj.orderBy = new List<Order> { new Order() { field = "value", descending = false } };
                    sqlObj.cmdParams = new Dictionary<string, object>();
                    sqlObj.cmdParams["@name1"] = "Value1";

                    {
                        var sql = api.GetSqlAsync(sqlObj).Result;
                        Console.WriteLine("sql:");
                        Console.WriteLine(sql);
                    }

                    {
                        using (var query = new query(serverUrl))
                        {
                            query.ExecuteAsync(sqlObj).Wait();
                            while (query.ReadAsync().Result)
                            {
                                Assert.AreEqual("Key1", query[0]); // value
                                Assert.AreEqual("Value1", query[1].ToString()); // name1
                                Assert.AreEqual("Value2", query[2].ToString()); // name2
                                Assert.AreEqual("", query[3]); // name3
                                Assert.AreEqual("", query[4]); // name4
                            }
                        }
                    }
                }

                // Clean up
                api.DeleteAsync(new Delete() { type = "ApiTable", values = new List<object> { "Key1", "Key2" } }).Wait();
                {
                    var gotten = api.GetAsync(new GetRequest() { type = "ApiTable", values = new List<object> { "Key1", "Key2"} }).Result;
                    Assert.AreEqual(2, gotten.metadata.Count);
                    Assert.AreEqual(null, gotten.metadata[0]);
                    Assert.AreEqual(null, gotten.metadata[1]);
                }

                api.DropAsync(new Drop() { type = "ApiTable" }).Wait();
                try
                {
                    var gotten = api.GetAsync(new GetRequest() { type = "ApiTable", values = new List<object> { "Key1", "Key2" } }).Result;
                    Assert.Fail();
                }
                catch { }

                api.ResetAsync(new Reset() { includeNameValues = false }).Wait();
                api.ResetAsync(new Reset() { includeNameValues = true }).Wait();
            }
        }
    }
}

