﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using Microsoft.VisualStudio.TestTools.UnitTesting;

namespace metastrings
{
    [TestClass]
    public class CsvTests
    {
        [TestMethod]
        public void TestCleanseCsv()
        {
            Assert.AreEqual("foo", CsvPrep.UnNullCsv("foo"));
            Assert.AreEqual(null, CsvPrep.UnNullCsv(CsvPrep.NullCellValue));

            Assert.AreEqual(CsvPrep.NullCellValue, CsvPrep.CleanseCsv(null));
            Assert.AreEqual("", CsvPrep.CleanseCsv(""));
            Assert.AreEqual("foo", CsvPrep.CleanseCsv("foo"));
            Assert.AreEqual("\"f\"\"o\"\"o\"", CsvPrep.CleanseCsv("f\"o\"o"));
            Assert.AreEqual("\"foo,\"", CsvPrep.CleanseCsv("foo,"));
            Assert.AreEqual("\"foo\"\",\"", CsvPrep.CleanseCsv("foo\","));
            Assert.AreEqual("\"\"\"foo\"\"bar,\"", CsvPrep.CleanseCsv("\"foo\"bar,"));

            Assert.AreEqual("\",bar\"", CsvPrep.CleanseCsv(",bar"));
        }

        [TestMethod]
        public void TestCreateCsv()
        {
            {
                var csv = CsvPrep.CreateCsv(new List<string>());
                Assert.AreEqual("", csv);
            }

            {
                var csv = CsvPrep.CreateCsv(new List<string>{ null });
                Assert.AreEqual(CsvPrep.NullCellValue, csv);
            }

            {
                var csv = CsvPrep.CreateCsv(new[] { "foo" });
                Assert.AreEqual("foo", csv);
            }

            {
                var csv = CsvPrep.CreateCsv(new[] { "foo", null, "bar" });
                Assert.AreEqual($"foo,{CsvPrep.NullCellValue},bar", csv);
            }

            {
                var csv = CsvPrep.CreateCsv(new[] { ",foo" });
                Assert.AreEqual("\",foo\"", csv);
            }

            {
                var csv = CsvPrep.CreateCsv(new[] { "foo," });
                Assert.AreEqual("\"foo,\"", csv);
            }

            {
                var csv = CsvPrep.CreateCsv(new[] { "foo", "bar" });
                Assert.AreEqual("foo,bar", csv);
            }

            {
                var csv = CsvPrep.CreateCsv(new[] { "foo", ",bar" });
                Assert.AreEqual("foo,\",bar\"", csv);
            }

            {
                var csv = CsvPrep.CreateCsv(new[] { "foo", "bar", "blet,monkey", "\"something\"" });
                Assert.AreEqual("foo,bar,\"blet,monkey\",\"\"\"something\"\"\"", csv);
            }
        }

        [TestMethod]
        public void TestParseCsv()
        {
            {
                string csv = CsvPrep.NullCellValue;
                using (var parser = new CsvParser(new StringReader(csv)))
                {
                    Assert.IsTrue(parser.Read());
                    Assert.AreEqual(1, parser.Count);
                    Assert.AreEqual(null, parser[0]);
                    Assert.IsFalse(parser.Read());
                }
            }

            {
                string csv = "";
                using (var parser = new CsvParser(new StringReader(csv)))
                {
                    Assert.IsFalse(parser.Read());
                }
            }

            {
                string csv = ",";
                using (var parser = new CsvParser(new StringReader(csv)))
                {
                    Assert.IsTrue(parser.Read());
                    Assert.AreEqual(2, parser.Count);
                    Assert.AreEqual("", parser[0]);
                    Assert.AreEqual("", parser[1]);

                    Assert.IsFalse(parser.Read());
                    Assert.IsFalse(parser.Read());
                }
            }

            {
                string csv = $"{CsvPrep.NullCellValue},{CsvPrep.NullCellValue}";
                using (var parser = new CsvParser(new StringReader(csv)))
                {
                    Assert.IsTrue(parser.Read());
                    Assert.AreEqual(2, parser.Count);
                    Assert.AreEqual(null, parser[0]);
                    Assert.AreEqual(null, parser[1]);
                }
            }

            {
                string csv = "foo";
                using (var parser = new CsvParser(new StringReader(csv)))
                {
                    Assert.IsTrue(parser.Read());
                    Assert.AreEqual(1, parser.Count);
                    Assert.AreEqual("foo", parser[0]);
                }
            }

            {
                string csv = "foo,";
                using (var parser = new CsvParser(new StringReader(csv)))
                {
                    Assert.IsTrue(parser.Read());
                    Assert.AreEqual(2, parser.Count);
                    Assert.AreEqual("foo", parser[0]);
                    Assert.AreEqual("", parser[1]);
                }
            }

            {
                string csv = "foo,,,,";
                using (var parser = new CsvParser(new StringReader(csv)))
                {
                    Assert.IsTrue(parser.Read());
                    Assert.AreEqual(5, parser.Count);
                    Assert.AreEqual("foo", parser[0]);
                    Assert.AreEqual("", parser[1]);
                    Assert.AreEqual("", parser[2]);
                    Assert.AreEqual("", parser[3]);
                    Assert.AreEqual("", parser[4]);
                }
            }

            {
                string csv = "\"foo\",,,,";
                using (var parser = new CsvParser(new StringReader(csv)))
                {
                    Assert.IsTrue(parser.Read());
                    Assert.AreEqual(5, parser.Count);
                    Assert.AreEqual("foo", parser[0]);
                    Assert.AreEqual("", parser[1]);
                    Assert.AreEqual("", parser[2]);
                    Assert.AreEqual("", parser[3]);
                    Assert.AreEqual("", parser[4]);
                }
            }

            {
                string csv = "foo,,bar";
                using (var parser = new CsvParser(new StringReader(csv)))
                {
                    Assert.IsTrue(parser.Read());
                    Assert.AreEqual(3, parser.Count);
                    Assert.AreEqual("foo", parser[0]);
                    Assert.AreEqual("", parser[1]);
                    Assert.AreEqual("bar", parser[2]);
                }
            }

            {
                string csv = "foo,bar";
                using (var parser = new CsvParser(new StringReader(csv)))
                {
                    Assert.IsTrue(parser.Read());
                    Assert.AreEqual(2, parser.Count);
                    Assert.AreEqual("foo", parser[0]);
                    Assert.AreEqual("bar", parser[1]);
                }
            }

            {
                string csv = "\"foo\"";
                using (var parser = new CsvParser(new StringReader(csv)))
                {
                    Assert.IsTrue(parser.Read());
                    Assert.AreEqual(1, parser.Count);
                    Assert.AreEqual("foo", parser[0]);
                }
            }

            {
                string csv = "\"fo\"\"o\"";
                using (var parser = new CsvParser(new StringReader(csv)))
                {
                    Assert.IsTrue(parser.Read());
                    Assert.AreEqual(1, parser.Count);
                    Assert.AreEqual("fo\"o", parser[0]);
                }
            }

            {
                string csv = "\"fo\"\"o\",blet,monkey\nsome,thing,else\n";
                using (var parser = new CsvParser(new StringReader(csv)))
                {
                    Assert.IsTrue(parser.Read());
                    Assert.AreEqual(3, parser.Count);
                    Assert.AreEqual("fo\"o", parser[0]);
                    Assert.AreEqual("blet", parser[1]);
                    Assert.AreEqual("monkey", parser[2]);

                    Assert.IsTrue(parser.Read());
                    Assert.AreEqual(3, parser.Count);
                    Assert.AreEqual("some", parser[0]);
                    Assert.AreEqual("thing", parser[1]);
                    Assert.AreEqual("else", parser[2]);
                }
            }
        }
    }
}
