﻿using System;
using System.Collections.Generic;
using System.Security.Cryptography;
using System.Text;
using System.IO;
using System.Net;
using System.Linq;

using Newtonsoft.Json;

namespace metastrings
{
    public static class Utils
    {
        public static string HashString(string input)
        {
            StringBuilder sb = new StringBuilder(64);
            using (var hasher = SHA256.Create())
            {
                byte[] hashBytes = hasher.ComputeHash(Encoding.Unicode.GetBytes(input));
                for (int i = 0; i < hashBytes.Length; i++)
                    sb.Append(hashBytes[i].ToString("x2"));
            }
            return sb.ToString();
        }

        public static T Deserialize<T>(string str)
        {
            using (TextReader txtReader = new StringReader(str))
            using (JsonReader jsReader = new JsonTextReader(txtReader))
                return sm_jsonSerializer.Deserialize<T>(jsReader);
        }

        public static string Serialize(object obj)
        {
            StringBuilder sb = new StringBuilder();
            using (TextWriter txtWriter = new StringWriter(sb))
                sm_jsonSerializer.Serialize(txtWriter, obj);
            return sb.ToString();
        }

        public static List<string> BatchUp(IEnumerable<string> pieces, int batchSize)
        {
            if (batchSize <= 0)
                throw new ArgumentException("batchSize must be > 0");

            var retVal = new List<string>();

            StringBuilder sb = new StringBuilder(batchSize);
            foreach (string str in pieces)
            {
                if (str.Length + sb.Length > batchSize)
                {
                    retVal.Add(sb.ToString());
                    sb.Clear();
                }

                sb.Append(str);
            }

            if (sb.Length > 0)
                retVal.Add(sb.ToString());

            return retVal;
        }

        public static List<List<T>> SplitUp<T>(IEnumerable<T> items, int batches)
        {
            if (batches <= 0)
                throw new ArgumentException("batches must be > 0");

            List<List<T>> retVal = new List<List<T>>(batches);

            if (batches == 1)
            {
                retVal.Add(new List<T>(items));
                return retVal;
            }

            int itemCount = items.Count();
            int batchSize = (int)Math.Round((double)itemCount / batches); // last batch can be fat

            for (int b = 0; b < batches; ++b)
            {
                var newList = new List<T>(batchSize);
                retVal.Add(newList);
            }

            int listIndex = 0;
            foreach (T t in items)
            {
                var curList = retVal[listIndex];
                curList.Add(t);

                if (curList.Count >= batchSize && listIndex != retVal.Count - 1)
                    ++listIndex;
            }

            return retVal;
        }

        public static Exception HandleException(Exception exp)
        {
            WebException webExp = null;
            if (exp is WebException)
                webExp = (WebException)exp;

            if (webExp == null || webExp.Response == null)
            {
                var innerExp = exp;
                while (innerExp.InnerException != null)
                    innerExp = innerExp.InnerException;

                string errorType = webExp == null ? "General" : "Network";
                return new MetaStringsException($"{errorType} Error: {innerExp.GetType().FullName}: {innerExp.Message}");
            }

            string responseString;
            using (var responseStream = webExp.Response.GetResponseStream())
            using (var sr = new StreamReader(responseStream, Encoding.UTF8))
                responseString = sr.ReadToEnd(); // can't do async in an exp handler

            return new MetaStringsException($"Server Error: {webExp.Status}: {responseString}");
        }

        private static JsonSerializer sm_jsonSerializer = new JsonSerializer();
    }
}
